from pathlib import Path
import random
import re
import pickle
from xml.etree import ElementTree
from lxml.etree import iterparse
from tqdm.auto import tqdm
import numpy as np
import torch
from torch.nn import functional as F
from torch.utils import data
from sklearn import metrics
from fastai.train import Learner, DataBunch
from .utils import text_to_id
from .defaults import CUT_LENGTH
from .models import Network

REMOVE_DOT_CHANCE = 0.5
LOWERCASE_START_CHANCE = 0.5
MIN_LENGTH = 300
MAX_LENGTH = 5000
N_CUTS = 2


def label_paragraph(tokenized_paragraph):
    full_text = ""
    labels = []

    for sentence in tokenized_paragraph:
        for i, token in enumerate(sentence):
            text = token.text
            whitespace = token.whitespace

            # add more whitespace with a small chance so that multiple whitespace is split
            # correctly too
            while True:
                if random.random() < 0.8:
                    break

                whitespace += " "

            text_to_append = text + whitespace

            if (
                text == "."
                and i == len(sentence) - 1
                and random.random() < REMOVE_DOT_CHANCE
            ):
                text_to_append = whitespace
                if len(text_to_append) > 0 and len(labels) > 1:
                    labels[-1][0] = 0.0

            if i == 0 and random.random() < LOWERCASE_START_CHANCE:
                text_to_append = text.lower() + whitespace

            for _ in range(len(text_to_append)):
                labels.append([0.0, 0.0])

            if len(labels) > 0:
                labels[-1][0] = 1.0

            full_text += text_to_append

        labels[-1][1] = 1.0

    return full_text, labels


def paragraph_to_text_and_labels(paragraph, n_cuts, cut_length):
    try:
        p_text, p_labels = label_paragraph(paragraph)
    except IndexError:
        print("Faulty paragraph:")
        print(paragraph)
        return [], []

    assert len(p_text) == len(p_labels)

    inputs = [[] for _ in range(n_cuts)]
    labels = [[] for _ in range(n_cuts)]

    for j in range(n_cuts):
        start = random.randint(0, len(p_text) + cut_length) - cut_length

        for k in range(cut_length):
            if start + k < 0 or start + k >= len(p_text):
                inputs[j].append(0)
                labels[j].append([0.0, 0.0])
            else:
                inputs[j].append(text_to_id(p_text[start + k]))
                labels[j].append(p_labels[start + k])

    return inputs, labels


def xml_to_paragraphs(
    corpus,
    max_n_paragraphs,
    min_length=MIN_LENGTH,
    max_length=MAX_LENGTH,
    store_path=None,
):
    def fast_iter(context):
        for event, elem in context:
            text = ElementTree.tostring(elem, encoding="utf8").decode("utf-8")
            text = re.sub(r"(<h>(.*?)<\/h>)", "\n", text)
            text = re.sub(r"<.*?>", "", text)
            text = text.strip()
            yield text

            # It's safe to call clear() here because no descendants will be
            # accessed
            elem.clear()
            # Also eliminate now-empty references from the root node to elem
            for ancestor in elem.xpath("ancestor-or-self::*"):
                while ancestor.getprevious() is not None:
                    parent = ancestor.getparent()

                    if parent is not None:
                        del parent[0]
                    else:
                        break

    i = 0
    paragraphs = []
    bar = tqdm(total=max_n_paragraphs)

    for p in fast_iter(iterparse(str(corpus), tag="p")):
        if not (min_length <= len(p) <= max_length):
            continue

        paragraphs.append(p)
        i += 1
        bar.update(1)

        if i >= max_n_paragraphs:
            break

    if store_path is not None:
        store_path = Path(store_path)
        store_path.parents[0].mkdir(exist_ok=True, parents=True)
        torch.save(paragraphs, store_path)

    return paragraphs


def prepare_tokenized_paragraphs(
    tokenized_paragraph_path,
    data_directory=None,
    subsample=None,
    remove_dot_chance=REMOVE_DOT_CHANCE,
    lowercase_start_chance=LOWERCASE_START_CHANCE,
    min_length=MIN_LENGTH,
    max_length=MAX_LENGTH,
    n_cuts=N_CUTS,
    cut_length=CUT_LENGTH,
):
    if data_directory is not None:
        data_directory = Path(data_directory)
        data_directory.mkdir(exist_ok=True, parents=True)

    all_sentences = []
    all_labels = []

    bar = tqdm()
    i = 0

    with open(tokenized_paragraph_path, "rb") as f:
        while True:
            try:
                paragraph = pickle.load(f)
                bar.update(1)

                i += 1

                if subsample is not None and i >= subsample:
                    break
            except EOFError:
                break

            text, labels = paragraph_to_text_and_labels(paragraph, n_cuts, cut_length)

            all_sentences.append(torch.tensor(text, dtype=torch.uint8))
            all_labels.append(torch.tensor(labels, dtype=torch.bool))

    all_sentences = torch.cat(all_sentences, 0)
    all_labels = torch.cat(all_labels, 0)

    if data_directory is not None:
        torch.save(all_sentences, data_directory / "all_sentences.pt")
        torch.save(all_labels, data_directory / "all_labels.pt")

    return all_sentences, all_labels


def loss(inputs, targets):
    weight = torch.tensor([1, 20]).view((1, 1, 2)).to(targets.device)
    return F.binary_cross_entropy_with_logits(inputs, targets.float(), weight=weight)


def train(
    sentences_train,
    labels_train,
    sentences_valid,
    labels_valid,
    batch_size=128,
    n_epochs=10,
):
    train_dataset = data.TensorDataset(sentences_train, labels_train)
    valid_dataset = data.TensorDataset(sentences_valid, labels_valid)

    model = Network()

    train_loader = data.DataLoader(
        train_dataset, batch_size=batch_size, shuffle=True, pin_memory=False
    )
    valid_loader = data.DataLoader(
        valid_dataset, batch_size=batch_size, shuffle=False, pin_memory=False
    )

    databunch = DataBunch(train_dl=train_loader, valid_dl=valid_loader)
    learn = Learner(databunch, model, loss_func=loss)

    if torch.cuda.is_available():
        learn = learn.to_fp16()

    learn.fit_one_cycle(n_epochs)

    return learn.model


def evaluate(model, sentences_valid, labels_valid, threshold=0.5, batch_size=1024):
    assert len(sentences_valid) == len(labels_valid)

    preds = np.zeros(labels_valid.shape)

    for i in tqdm(range((len(sentences_valid) // batch_size) + 1)):
        idx = slice(i * batch_size, (i + 1) * batch_size)

        if next(model.parameters()).is_cuda:
            out = model(sentences_valid[idx].cuda())
        else:
            out = model(sentences_valid[idx])

        preds[idx] = out.detach().cpu().numpy()

    preds = preds.reshape((-1, 2)) > threshold
    labels_valid = labels_valid.cpu().numpy().reshape((-1, 2))

    for (i, name) in zip([0, 1], ["Tokenize", "Sentencize"]):
        print(f"Target: {name} \n")

        for (metric, name) in [
            [metrics.f1_score, "F1"],
            [metrics.precision_score, "Precision"],
            [metrics.recall_score, "Recall"],
        ]:
            print(f"{name}:", metric(preds[:, i], labels_valid[:, i]))

        print("\n\n")



__version__ = "0.2.2"
__all__ = ["NNSplit"]

import numpy as np
import torch
from .defaults import CUT_LENGTH, DEVICE
from .utils import load_provided_model, text_to_id
from .tokenizer import Tokenizer, Token


def _get_token(text):
    last_char_index = 0
    index = len(text) - 1

    while index >= 0:
        if not text[index].isspace():
            last_char_index = index + 1
            break

        index -= 1

    return Token(text[:last_char_index], text[last_char_index:])


class NNSplit(Tokenizer):
    """
    An NNSplit sentencizer and tokenizer.

    Parameters:
        model_or_model_name (str or torch.nn.Module or torch.jit.TracedModule):
            If str (either `de` or `en`) a pretrained model will be loaded.
            If torch.nn.Module or torch.jit.TracedModule, the given module will be used instead.
        threshold (float, 0 < x < 1): Cutoff above which predictions will be considered as 1.
        stride (int):
            How much to move the window after each prediction.
            Comparable to stride in a 1d convolution.
        cut_length (int):
            The number of characters in each cut.
        device (torch.device):
            On which device to run the model.
    """

    def __init__(
        self,
        model_or_model_name,
        threshold=0.1,
        stride=CUT_LENGTH // 2,
        cut_length=CUT_LENGTH,
        device=DEVICE,
    ):
        self.threshold = threshold
        self.stride = stride
        self.cut_length = cut_length
        self.device = device
        self.padding = 5

        if isinstance(model_or_model_name, (torch.nn.Module, torch.jit.TracedModule)):
            self.model = model_or_model_name
        else:
            self.model = load_provided_model(model_or_model_name, device)

    def _get_raw_preds(self, texts, batch_size):
        all_inputs = []
        all_idx = []
        n_cuts_per_text = []

        for text in texts:
            inputs = (
                [0] * self.padding + [text_to_id(x) for x in text] + [0] * self.padding
            )

            while len(inputs) < self.cut_length:
                inputs.append(0)

            start = 0
            end = -1
            i = 0

            # split the inputs into partially overlapping slices which all have the same length
            # allows efficient batching
            while end != len(inputs):
                end = min(start + self.cut_length, len(inputs))
                start = end - self.cut_length

                idx = slice(start, end)
                all_inputs.append(inputs[idx])
                all_idx.append(idx)

                start += self.stride
                i += 1

            n_cuts_per_text.append(i)

        batched_inputs = torch.tensor(all_inputs, dtype=torch.int64)
        preds = np.zeros((len(batched_inputs), self.cut_length, 2), dtype=np.float32)

        for start in range(0, len(batched_inputs), batch_size):
            end = start + batch_size
            preds[start:end] = torch.sigmoid(
                self.model(batched_inputs[start:end].to(self.device))
                .detach()
                .float()
                .cpu()
            ).numpy()

        return preds, all_idx, n_cuts_per_text

    def _average_preds(self, texts, preds, all_idx, n_cuts_per_text):
        all_avg_preds = [
            np.zeros((self.padding * 2 + len(text), 3), dtype=np.float32)
            for text in texts
        ]

        current_text = 0
        current_i = 0
        for pred, idx in zip(preds, all_idx):
            current_preds = all_avg_preds[current_text]

            # add current slice to average preds
            current_preds[idx, :2] += pred[: len(current_preds)]
            current_preds[idx, 2] += 1

            current_i += 1

            if current_i == n_cuts_per_text[current_text]:
                # divide predictions by number of predictions so they are on the same scale
                all_avg_preds[current_text] = (
                    current_preds[self.padding : -self.padding, :2]
                    / current_preds[self.padding : -self.padding, [2]]
                )

                current_text += 1
                current_i = 0

        return all_avg_preds

    def _split_text_from_preds(self, texts, all_avg_preds):
        tokenized_texts = []

        for text, avg_preds in zip(texts, all_avg_preds):
            avg_preds = avg_preds > self.threshold
            # for better handling with np.where, so that each index is only interated over once
            avg_preds[:, 0] &= ~avg_preds[:, 1]

            sentences = []
            tokens = []

            prev_index = 0
            index = 0

            for index, kind in zip(*np.where(avg_preds)):
                index = index + 1

                tokens.append(_get_token(text[prev_index:index]))

                if kind == 1:
                    sentences.append(tokens)
                    tokens = []

                prev_index = index

            if index != len(text):
                tokens.append(_get_token(text[index:]))

            if len(tokens) > 0:
                sentences.append(tokens)

            tokenized_texts.append(sentences)

        return tokenized_texts

    def split(self, texts, batch_size=128):
        """
        Split texts into sentences and tokens.

        Parameters:
            texts (List[str]):
                A list of texts to split.
                Passing multiple texts at once allows for parallelization of the model.
            batch_size (int, Optional):
                Batch size with which cuts are processed by the model.

        Returns:
            A list with the same length as `texts`.
            - Each element is a list of sentences.
            - Each sentence is a list of tokens.
            - Each token is a namedtuple with `text` and `whitespace`.
        """
        preds, all_idx, n_cuts_per_text = self._get_raw_preds(texts, batch_size)
        all_avg_preds = self._average_preds(texts, preds, all_idx, n_cuts_per_text)

        return self._split_text_from_preds(texts, all_avg_preds)



import numpy as np
from torch import nn


def _freeze_bias(lstm):
    for name, param in lstm.named_parameters():
        if name.startswith("bias"):
            param.requires_grad = False
            param[:] = 0


class Network(nn.Module):
    def __init__(self):
        super().__init__()
        self.embedding = nn.Embedding(127 + 2, 32)
        self.lstm1 = nn.LSTM(32, 128, bidirectional=True, batch_first=True)
        _freeze_bias(self.lstm1)
        self.lstm2 = nn.LSTM(256, 64, bidirectional=True, batch_first=True)
        _freeze_bias(self.lstm2)
        self.out = nn.Linear(128, 2)

    def get_keras_equivalent(self):
        from tensorflow.keras import layers, models

        k_model = models.Sequential()
        k_model.add(layers.Input(shape=(None,)))

        k_model.add(layers.Embedding(127 + 2, 32))
        k_model.layers[-1].set_weights([self.embedding.weight.detach().cpu().numpy()])

        k_model.add(
            layers.Bidirectional(
                layers.LSTM(128, return_sequences=True, use_bias=False)
            )
        )
        k_model.layers[-1].set_weights(
            [
                np.transpose(x.detach().cpu().numpy())
                for name, x in self.lstm1.named_parameters()
                if not name.startswith("bias")
            ]
        )

        k_model.add(
            layers.Bidirectional(layers.LSTM(64, return_sequences=True, use_bias=False))
        )
        k_model.layers[-1].set_weights(
            [
                np.transpose(x.detach().cpu().numpy())
                for name, x in self.lstm2.named_parameters()
                if not name.startswith("bias")
            ]
        )

        k_model.add(layers.Dense(2))
        k_model.layers[-1].set_weights(
            [np.transpose(x.detach().cpu().numpy()) for x in self.out.parameters()]
        )
        return k_model

    def forward(self, x):
        h = self.embedding(x.long())
        h, _ = self.lstm1(h)
        h, _ = self.lstm2(h)
        h = self.out(h)
        return h



import io
import logging
from pathlib import Path
import pkgutil
import torch
from torch import nn
import numpy as np
from .defaults import DEVICE


def text_to_id(char):
    x = ord(char)
    return x + 2 if x <= 127 else 1


def id_to_text(x):
    return chr(x - 2) if (x - 2) <= 127 and x > 1 else "X"


def print_sentence_and_label(sentence, label):
    for char_id, label in zip(sentence, label):
        print(id_to_text(char_id), int(label[0]), int(label[1]))


def label_tokens(tokenized_p):
    text = ""
    token_labels, sentence_labels = [], []

    for sentence in tokenized_p:
        for token in sentence:
            raw_token = token.text + token.whitespace

            text += raw_token
            token_labels += [0] * len(raw_token)
            sentence_labels += [0] * len(raw_token)

            token_labels[-1] = 1.0

        sentence_labels[-1] = 1.0

    return text, np.stack([token_labels, sentence_labels], -1)


def store_model(model, store_directory):
    store_directory = Path(store_directory)
    store_directory.mkdir(exist_ok=True, parents=True)

    sample = torch.zeros([1, 100])
    # model is trained with fp16, so it can be safely quantized to 16 bit
    # CPU model is quantized to 8 bit, with minimal loss in accuracy
    # according to tests in train.ipynb
    quantized_model = torch.quantization.quantize_dynamic(
        model.float().cpu(), {nn.LSTM, nn.Linear}, dtype=torch.qint8
    )
    traced = torch.jit.trace(quantized_model, sample)
    traced.save(str(store_directory / "ts_cpu.pt"))

    if torch.cuda.is_available():
        traced = torch.jit.trace(model.half().cuda(), sample.cuda())
        traced.save(str(store_directory / "ts_cuda.pt"))
    else:
        logging.warn(
            "CUDA is not available. CUDA version of model could not be stored."
        )

    import tensorflowjs as tfjs  # noqa: F401

    tfjs.converters.save_keras_model(
        model.get_keras_equivalent(),
        str(store_directory / "tfjs_model"),
        quantization_dtype=np.uint8,
    )


def _get_filename(device):
    filename = (
        "ts_cpu.pt"
        if device == torch.device("cpu") or not torch.cuda.is_available()
        else "ts_cuda.pt"
    )
    return filename


def load_provided_model(name, device=DEVICE):
    bin_data = pkgutil.get_data(__package__, f"data/{name}/{_get_filename(device)}")
    buffer = io.BytesIO(bin_data)

    return torch.jit.load(buffer)


def load_model(store_directory, device=DEVICE):
    full_path = Path(store_directory) / _get_filename(device)

    return torch.jit.load(str(full_path))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nnsplit']

package_data = \
{'': ['*'],
 'nnsplit': ['data/*',
             'data/de/*',
             'data/de/tfjs_model/*',
             'data/en/*',
             'data/en/tfjs_model/*']}

install_requires = \
['numpy>=1,<2', 'torch>=1.4,<2.0']

setup_kwargs = {
    'name': 'nnsplit',
    'version': '0.2.2',
    'description': 'Fast, robust sentence splitting with bindings for Python, Rust and Javascript.',
    'long_description': '# NNSplit Python Bindings\n\n![PyPI](https://img.shields.io/pypi/v/nnsplit)\n![CI](https://github.com/bminixhofer/nnsplit/workflows/CI/badge.svg)\n![License](https://img.shields.io/github/license/bminixhofer/nnsplit)\n\nFast, robust sentence splitting with bindings for Python, Rust and Javascript and pretrained models for English and German.\n\n## Installation\n\nNNSplit has PyTorch as the only dependency.\n\nInstall it with pip: `pip install nnsplit`\n\n## Usage\n\n```python\n>>> from nnsplit import NNSplit\n>>> splitter = NNSplit("en")\n# NNSplit does not depend on proper punctuation and casing to split sentences\n>>> splitter.split(["This is a test This is another test."])\n[[[Token(text=\'This\', whitespace=\' \'),\n   Token(text=\'is\', whitespace=\' \'),\n   Token(text=\'a\', whitespace=\' \'),\n   Token(text=\'test\', whitespace=\' \')],\n  [Token(text=\'This\', whitespace=\' \'),\n   Token(text=\'is\', whitespace=\' \'),\n   Token(text=\'another\', whitespace=\' \'),\n   Token(text=\'test\', whitespace=\'\'),\n   Token(text=\'.\', whitespace=\'\')]]]\n```\n\nModels for German (`NNSplit("de")`) and English (`NNSplit("en")`) come prepackaged with NNSplit. Alternatively, you can also load your own model:\n\n```python\nimport torch\nmodel = torch.jit.load("/path/to/your/model.pt") # a regular nn.Module works too\n\nsplitter = NNSplit(model)\n```\n\nSee `train.ipynb` for the code used to train the pretrained models.\n\n## Development\n\nNNSplit uses [Poetry](https://python-poetry.org/) for dependency management. I made a small `Makefile` to automate some steps. Take a look at the `Makefile` and run `make install`, `make build`, `make test` to install, build and test the library, respectively.',
    'author': 'Benjamin Minixhofer',
    'author_email': 'bminixhofer@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



__all__ = ["Token", "Tokenizer", "SoMaJoTokenizer"]

from abc import ABC, abstractmethod
from collections import namedtuple

Token = namedtuple("Token", ["text", "whitespace"])


class Tokenizer(ABC):
    @abstractmethod
    def split(self, texts):
        """
        `split` returns an iterable over texts. Every text is an iterable over sentences.
        Every sentence is an iterable over `Token` namedtuple instances.
        """
        pass


class SoMaJoTokenizer(Tokenizer):
    def __init__(self, language, processes=None):
        from somajo import SoMaJo

        tokenizer_type = {"de": "de_CMC", "en": "en_PTB"}[language]
        self.tokenizer = SoMaJo(
            tokenizer_type, split_camel_case=True, split_sentences=True
        )

    def _tokenize_text(self, text):
        sentences = []

        if len(text) == 0:
            return sentences

        for sentence in self.tokenizer.tokenize_text([text]):
            sentences.append(
                [
                    Token(token.text, " " if token.space_after else "")
                    for token in sentence
                ]
            )

        if not text[-1].isspace():
            sentences[-1][-1] = Token(sentences[-1][-1].text, "")

        return sentences

    def split(self, texts, verbose=False):
        bar = None
        if verbose:
            from tqdm.auto import tqdm

            bar = tqdm(total=len(texts))

        # pool.imap leaks memory for some reason
        for sentences in map(self._tokenize_text, texts):
            yield sentences

            if verbose:
                bar.update(1)
